# BioPandas
# Author: Sebastian Raschka <mail@sebastianraschka.com>
# License: BSD 3 clause
# Project Website: http://rasbt.github.io/biopandas/
# Code Repository: https://github.com/rasbt/biopandas

from biopandas.pdb import PandasPdb
import warnings
import pandas as pd
import os


TESTDATA_FILENAME = os.path.join(os.path.dirname(__file__), 'data', '3eiy.pdb')
TESTDATA_FILENAME2 = os.path.join(os.path.dirname(__file__), 'data',
                                  '4eiy_anisouchunk.pdb')
OUTFILE = os.path.join(os.path.dirname(__file__), 'data', 'tmp.pdb')
OUTFILE_GZ = os.path.join(os.path.dirname(__file__), 'data', 'tmp.pdb.gz')

hetatm = ''
with open(TESTDATA_FILENAME, 'r') as f:
    for line in f:
        if line.startswith('HETATM'):
            hetatm += line

with open(TESTDATA_FILENAME2, 'r') as f:
    four_eiy = f.read()


def test_defaults():
    ppdb = PandasPdb()
    ppdb.read_pdb(TESTDATA_FILENAME)
    ppdb.to_pdb(path=OUTFILE, records=None)
    with open(TESTDATA_FILENAME, 'r') as f:
        f1 = f.read()
    with open(OUTFILE, 'r') as f:
        f2 = f.read()
    assert f1 == f2
    os.remove(OUTFILE)


def test_nonexpected_column():
    ppdb = PandasPdb()
    ppdb.read_pdb(TESTDATA_FILENAME)
    ppdb.df['HETATM']['test'] = pd.Series('test',
                                          index=ppdb.df['HETATM'].index)
    with warnings.catch_warnings(record=True) as w:
        ppdb.to_pdb(path=OUTFILE, records=['HETATM'])
    with open(OUTFILE, 'r') as f:
        f1 = f.read()
    os.remove(OUTFILE)
    assert f1 == hetatm


def test_records():
    """Test private _read_pdb."""
    ppdb = PandasPdb()
    ppdb.read_pdb(TESTDATA_FILENAME)
    ppdb.to_pdb(path=OUTFILE, records=['HETATM'])
    with open(OUTFILE, 'r') as f:
        f1 = f.read()
    os.remove(OUTFILE)
    assert f1 == hetatm


def test_anisou():
    """Test writing ANISOU entries."""
    ppdb = PandasPdb()
    ppdb.read_pdb(TESTDATA_FILENAME2)
    ppdb.to_pdb(path=OUTFILE, records=None)
    with open(OUTFILE, 'r') as f:
        f1 = f.read()
    os.remove(OUTFILE)
    assert f1 == four_eiy



# BioPandas
# Author: Sebastian Raschka <mail@sebastianraschka.com>
# License: BSD 3 clause
# Project Website: http://rasbt.github.io/biopandas/
# Code Repository: https://github.com/rasbt/biopandas


from biopandas.pdb import PandasPdb
import os
import numpy as np
import pandas as pd
from nose.tools import raises
from biopandas.testutils import assert_raises
try:
    from urllib.request import urlopen
    from urllib.error import HTTPError, URLError
except ImportError:
    from urllib2 import urlopen, HTTPError, URLError  # Python 2.7 compatib


TESTDATA_FILENAME = os.path.join(os.path.dirname(__file__), 'data', '3eiy.pdb')
TESTDATA_FILENAME2 = os.path.join(os.path.dirname(__file__), 'data',
                                  '4eiy_anisouchunk.pdb')
TESTDATA_FILENAME_GZ = os.path.join(os.path.dirname(__file__), 'data',
                                    '3eiy.pdb.gz')

ATOM_DF_COLUMNS = ['record_name', 'atom_number', 'blank_1',
                   'atom_name', 'alt_loc', 'residue_name',
                   'blank_2', 'chain_id', 'residue_number',
                   'insertion', 'blank_3',
                   'x_coord', 'y_coord', 'z_coord',
                   'occupancy', 'b_factor', 'blank_4',
                   'segment_id', 'element_symbol',
                   'charge', 'line_idx']

ANISOU_DF_COLUMNS = ['record_name', 'atom_number', 'blank_1',
                     'atom_name', 'alt_loc', 'residue_name',
                     'blank_2', 'chain_id', 'residue_number',
                     'insertion', 'blank_3',
                     'U(1,1)', 'U(2,2)', 'U(3,3)',
                     'U(1,2)', 'U(1,3)', 'U(2,3)',
                     'blank_4', 'element_symbol',
                     'charge', 'line_idx']

with open(TESTDATA_FILENAME, 'r') as f:
    three_eiy = f.read()

with open(TESTDATA_FILENAME2, 'r') as f:
    four_eiy = f.read()


def test__read_pdb():
    """Test private _read_pdb"""
    ppdb = PandasPdb()
    path, txt = ppdb._read_pdb(TESTDATA_FILENAME)
    print(txt)
    assert txt == three_eiy


def test__read_pdb_raises():
    """Test private _read_pdb:
    Test if ValueError is raised for wrong file formats."""

    expect = ('Wrong file format; allowed file formats'
              ' are .pdb and .pdb.gz.')

    def run_code_1():
        PandasPdb()._read_pdb("protein.mol2")

    assert_raises(ValueError,
                  expect,
                  run_code_1)

    def run_code_2():
        PandasPdb()._read_pdb("protein.mol2.gz")

    assert_raises(ValueError,
                  expect,
                  run_code_2)


def test_fetch_pdb():
    """Test fetch_pdb"""

    try:
        ppdb = PandasPdb()
        url, txt = ppdb._fetch_pdb('3eiy')
    except HTTPError:
        url, txt = None, None
    except ConnectionResetError:
        url, txt = None, None

    if txt:  # skip if PDB down
        txt[:100] == three_eiy[:100]
        ppdb.fetch_pdb('3eiy')
        assert ppdb.pdb_text == txt
        assert ppdb.pdb_path == 'https://files.rcsb.org/download/3eiy.pdb'


def test__read_pdb_gz():
    """Test public _read_pdb with gzip files"""
    ppdb = PandasPdb()
    path, txt = ppdb._read_pdb(TESTDATA_FILENAME_GZ)
    assert txt == three_eiy


def test__construct_df():
    """Test pandas dataframe construction"""
    ppdb = PandasPdb()
    dfs = ppdb._construct_df(three_eiy.splitlines())
    assert set(dfs.keys()) == {'OTHERS', 'ATOM', 'ANISOU', 'HETATM'}
    assert set(dfs['ATOM'].columns) == set(ATOM_DF_COLUMNS)
    assert set(dfs['HETATM'].columns) == set(ATOM_DF_COLUMNS)
    assert set(dfs['ANISOU'].columns) == set(ANISOU_DF_COLUMNS)
    exp = pd.Series(np.array(['ATOM', 1, '', 'N', '', 'SER',
                              '', 'A', 2, '', '', 2.527, 54.656, -1.667, 1.0,
                              52.73, '', '', 'N', None, 609]),
                    index=['record_name', 'atom_number', 'blank_1',
                           'atom_name', 'alt_loc', 'residue_name',
                           'blank_2', 'chain_id', 'residue_number',
                           'insertion', 'blank_3',
                           'x_coord', 'y_coord', 'z_coord',
                           'occupancy', 'b_factor', 'blank_4',
                           'segment_id', 'element_symbol',
                           'charge', 'line_idx'])
    assert exp.equals(dfs['ATOM'].loc[0, :])


def test_read_pdb():
    """Test public read_pdb"""
    ppdb = PandasPdb()
    ppdb.read_pdb(TESTDATA_FILENAME)
    assert ppdb.pdb_text == three_eiy
    assert ppdb.code == '3eiy', ppdb.code
    assert ppdb.pdb_path == TESTDATA_FILENAME


def test_read_pdb_from_list():
    """Test public read_pdb_from_list"""

    for pdb_text, code in zip([three_eiy, four_eiy], ['3eiy', '4eiy']):
        ppdb = PandasPdb()
        ppdb.read_pdb_from_list(pdb_text.splitlines(True))
        assert ppdb.pdb_text == pdb_text
        assert ppdb.code == code
        assert ppdb.pdb_path == ''


def test_anisou_input_handling():
    """Test public read_pdb"""
    ppdb = PandasPdb()
    ppdb.read_pdb(TESTDATA_FILENAME2)
    assert ppdb.pdb_text == four_eiy
    assert ppdb.code == '4eiy', ppdb.code


@raises(AttributeError)
def test_get_exceptions():
    ppdb = PandasPdb()
    ppdb.read_pdb(TESTDATA_FILENAME)
    ppdb.get('main-chai')


def test_get_all():
    ppdb = PandasPdb()
    ppdb.read_pdb(TESTDATA_FILENAME)
    for i in ['c-alpha', 'hydrogen', 'main chain']:
        ppdb.get(i)


def test_get_df():
    ppdb = PandasPdb()
    ppdb.read_pdb(TESTDATA_FILENAME)

    shape = ppdb.get('c-alpha').shape
    assert shape == (174, 21), shape

    shape = ppdb.get('hydrogen', invert=True, records=('ATOM',)).shape
    assert shape == (1330, 21), shape

    # deprecated use of string
    shape = ppdb.get('hydrogen', invert=True, records='ATOM').shape
    assert shape == (1330, 21), shape

    shape = ppdb.get('hydrogen').shape
    assert shape == (0, 21), shape

    shape = ppdb.get('main chain', records=('ATOM',)).shape
    assert shape == (696, 21), shape

    shape = ppdb.get('heavy', records=('ATOM',)).shape
    assert shape == (1330, 21), shape

    shape = ppdb.get('carbon', records=('ATOM',)).shape
    assert shape == (473, 21), shape



""" Class for working with PDB files"""

# BioPandas
# Author: Sebastian Raschka <mail@sebastianraschka.com>
# License: BSD 3 clause
# Project Website: http://rasbt.github.io/biopandas/
# Code Repository: https://github.com/rasbt/biopandas

import pandas as pd
import numpy as np
import sys
import gzip
from warnings import warn
try:
    from urllib.request import urlopen
    from urllib.error import HTTPError, URLError
except ImportError:
    from urllib2 import urlopen, HTTPError, URLError  # Python 2.7 compatible
from .engines import pdb_records
from .engines import pdb_df_columns
from .engines import amino3to1dict
import warnings
from distutils.version import LooseVersion


pd_version = LooseVersion(pd.__version__)


class PandasPdb(object):
    """
    Object for working with Protein Databank structure files.

    Attributes
    ----------
    df : dict
        Dictionary storing pandas DataFrames for PDB record sections.
        The dictionary keys are {'ATOM', 'HETATM', 'ANISOU', 'OTHERS'}
        where 'OTHERS' contains all entries that are not parsed as
        'ATOM', 'HETATM', or 'ANISOU'.

    pdb_text : str
        PDB file contents in raw text format.

    pdb_path : str
        Location of the PDB file that was read in via `read_pdb`
        or URL of the page where the PDB content was fetched from
        if `fetch_pdb` was called.

    header : str
        PDB file description.

    code : str
        PDB code

    """
    def __init__(self):
        self._df = {}
        self.pdb_text = ''
        self.header = ''
        self.code = ''
        self._get_dict = {}
        self.pdb_path = ''

    @property
    def df(self):
        """Acccess dictionary of pandas DataFrames for PDB record sections."""
        return self._df

    @df.setter
    def df(self, value):
        """Assign a new value to the pandas DataFrame"""
        raise AttributeError('Please use `PandasPdb._df = ... ` instead\n'
                             'of `PandasPdb.df = ... ` if you are sure that\n'
                             'you want to overwrite the `df` attribute.')
        # self._df = value

    def read_pdb(self, path):
        """Read PDB files (unzipped or gzipped) from local drive

        Attributes
        ----------
        path : str
            Path to the PDB file in .pdb format or gzipped format (.pdb.gz).

        Returns
        ---------
        self

        """
        self.pdb_path, self.pdb_text = self._read_pdb(path=path)
        self._df = self._construct_df(pdb_lines=self.pdb_text.splitlines(True))
        self.header, self.code = self._parse_header_code()
        return self

    def read_pdb_from_list(self, pdb_lines):
        """Reads PDB file from a list into DataFrames

        Attributes
        ----------
        pdb_lines : list
            A list of lines containing the pdb file contents.

        Returns
        ---------
        self

        """
        self.pdb_text = ''.join(pdb_lines)
        self._df = self._construct_df(pdb_lines)
        self.header, self.code = self._parse_header_code()
        return self

    def fetch_pdb(self, pdb_code):
        """Fetches PDB file contents from the Protein Databank at rcsb.org.

        Parameters
        ----------
        pdb_code : str
            A 4-letter PDB code, e.g., "3eiy".

        Returns
        ---------
        self

        """
        self.pdb_path, self.pdb_text = self._fetch_pdb(pdb_code)
        self._df = self._construct_df(pdb_lines=self.pdb_text.splitlines(True))
        return self

    def get(self, s, df=None, invert=False, records=('ATOM', 'HETATM')):
        """Filter PDB DataFrames by properties

        Parameters
        ----------
        s : str  in {'main chain', 'hydrogen', 'c-alpha', 'heavy'}
            String to specify which entries to return.

        df : pandas.DataFrame, default: None
            Optional DataFrame to perform the filter operation on.
            If df=None, filters on self.df['ATOM'].

        invert : bool, default: True
            Inverts the search query. For example if s='hydrogen' and
            invert=True, all but hydrogen entries are returned.

        records : iterable, default: ('ATOM', 'HETATM')
            Specify which record sections to consider. For example, to consider
            both protein and ligand atoms, set `records=('ATOM', 'HETATM')`.
            This setting is ignored if `df` is not set to None.
            For downward compatibility, a string argument is still supported
            but deprecated and will be removed in future versions.

        Returns
        --------
        df : pandas.DataFrame
            Returns a DataFrame view on the filtered entries.

        """
        if isinstance(records, str):
            warnings.warn('Using a string as `records` argument is '
                          'deprecated and will not be supported in future'
                          ' versions. Please use a tuple or'
                          ' other iterable instead', DeprecationWarning)
            records = (records,)

        if not self._get_dict:
            self._get_dict = self._init_get_dict()
        if s not in self._get_dict.keys():
            raise AttributeError('s must be in %s' % self._get_dict.keys())
        if not df:
            df = pd.concat(objs=[self.df[i] for i in records])
        return self._get_dict[s](df, invert=invert)

    def impute_element(self, records=('ATOM', 'HETATM'), inplace=False):
        """Impute element_symbol from atom_name section.

        Parameters
        ----------
        records : iterable, default: ('ATOM', 'HETATM')
            Coordinate sections for which the element symbols should be
            imputed.

        inplace : bool, (default: False
            Performs the operation in-place if True and returns a copy of the
            PDB DataFrame otherwise.

        Returns
        ---------
        DataFrame

        """
        if inplace:
            t = self.df
        else:
            t = self.df.copy()
            for d in self.df:
                t[d] = self.df[d].copy()

        for sec in records:
            t[sec]['element_symbol'] = \
                t[sec][['atom_name', 'element_symbol']].\
                apply(lambda x: x[0][1]
                      if len(x[1]) == 3
                      else x[0][0], axis=1)
        return t

    @staticmethod
    def rmsd(df1, df2, s=None, invert=False):
        """Compute the Root Mean Square Deviation between molecules.

        Parameters
        ----------
        df1 : pandas.DataFrame
            DataFrame with HETATM, ATOM, and/or ANISOU entries.

        df2 : pandas.DataFrame
            Second DataFrame for RMSD computation against df1. Must have the
            same number of entries as df1.

        s : {'main chain', 'hydrogen', 'c-alpha', 'heavy', 'carbon'} or None,
            default: None
            String to specify which entries to consider. If None, considers
            all atoms for comparison.

        invert : bool, default: False
            Inverts the string query if true. For example, the setting
            `s='hydrogen', invert=True` computes the RMSD based on all
            but hydrogen atoms.

        Returns
        ---------
        rmsd : float
            Root Mean Square Deviation between df1 and df2

        """
        if df1.shape[0] != df2.shape[0]:
            raise AttributeError('DataFrames have unequal lengths')
        get_dict = PandasPdb._init_get_dict()
        if s:
            if s not in get_dict.keys():
                raise AttributeError('s must be in '
                                     '%s or None' % get_dict.keys())
            df1 = get_dict[s](df1, invert=invert)
            df2 = get_dict[s](df2, invert=invert)

        total = ((df1['x_coord'].values - df2['x_coord'].values)**2 +
                 (df1['y_coord'].values - df2['y_coord'].values)**2 +
                 (df1['z_coord'].values - df2['z_coord'].values)**2)
        rmsd = round((total.sum() / df1.shape[0])**0.5, 4)
        return rmsd

    @staticmethod
    def _init_get_dict():
        """Initialize dictionary for filter operations."""
        get_dict = {'main chain': PandasPdb._get_mainchain,
                    'hydrogen': PandasPdb._get_hydrogen,
                    'c-alpha': PandasPdb._get_calpha,
                    'carbon': PandasPdb._get_carbon,
                    'heavy': PandasPdb._get_heavy}
        return get_dict

    @staticmethod
    def _read_pdb(path):
        """Read PDB file from local drive."""
        r_mode = 'r'
        if path.endswith('.pdb'):
            openf = open
        elif path.endswith('pdb.gz'):
            r_mode = 'rb'
            openf = gzip.open
        else:
            raise ValueError(
                'Wrong file format; allowed file formats are .pdb and .pdb.gz.'
            )

        with openf(path, r_mode) as f:
            txt = f.read()

        if path.endswith('.gz'):
            if sys.version_info[0] >= 3:
                txt = txt.decode('utf-8')
            else:
                txt = txt.encode('ascii')
        return path, txt

    @staticmethod
    def _fetch_pdb(pdb_code):
        """Load PDB file from rcsb.org."""
        txt = None
        url = 'https://files.rcsb.org/download/%s.pdb' % pdb_code.lower()
        try:
            response = urlopen(url)
            txt = response.read()
            if sys.version_info[0] >= 3:
                txt = txt.decode('utf-8')
            else:
                txt = txt.encode('ascii')
        except HTTPError as e:
            print('HTTP Error %s' % e.code)
        except URLError as e:
            print('URL Error %s' % e.args)
        return url, txt

    def _parse_header_code(self):
        """Extract header information and PDB code."""
        code, header = '', ''
        if 'OTHERS' in self.df:

            header = (self.df['OTHERS'][self.df['OTHERS']['record_name'] ==
                      'HEADER'])
            if not header.empty:
                header = header['entry'].values[0]
                s = header.split()
                if s:
                    code = s[-1].lower()
        return header, code

    @staticmethod
    def _get_mainchain(df, invert):
        """Return only main chain atom entries from a DataFrame"""
        if invert:
            mc = df[(df['atom_name'] != 'C') &
                    (df['atom_name'] != 'O') &
                    (df['atom_name'] != 'N') &
                    (df['atom_name'] != 'CA')]
        else:
            mc = df[(df['atom_name'] == 'C') |
                    (df['atom_name'] == 'O') |
                    (df['atom_name'] == 'N') |
                    (df['atom_name'] == 'CA')]
        return mc

    @staticmethod
    def _get_hydrogen(df, invert):
        """Return only hydrogen atom entries from a DataFrame"""
        if invert:
            return df[(df['element_symbol'] != 'H')]
        else:
            return df[(df['element_symbol'] == 'H')]

    @staticmethod
    def _get_heavy(df, invert):
        """Return only heavy atom entries from a DataFrame"""
        if invert:
            return df[df['element_symbol'] == 'H']
        else:
            return df[df['element_symbol'] != 'H']

    @staticmethod
    def _get_calpha(df, invert):
        """Return c-alpha atom entries from a DataFrame"""
        if invert:
            return df[df['atom_name'] != 'CA']
        else:
            return df[df['atom_name'] == 'CA']

    @staticmethod
    def _get_carbon(df, invert):
        """Return c-alpha atom entries from a DataFrame"""
        if invert:
            return df[df['element_symbol'] == 'C']
        else:
            return df[df['element_symbol'] != 'C']

    @staticmethod
    def _construct_df(pdb_lines):
        """Construct DataFrames from list of PDB lines."""
        valids = tuple(pdb_records.keys())
        line_lists = {r: [] for r in valids}
        line_lists['OTHERS'] = []
        for line_num, line in enumerate(pdb_lines):
            if line.strip():
                if line.startswith(valids):
                    record = line[:6].rstrip()
                    line_ele = ['' for _ in range(len(
                        pdb_records[record]) + 1)]
                    for idx, ele in enumerate(pdb_records[record]):
                        line_ele[idx] = (line[ele['line'][0]:ele['line'][1]]
                                         .strip())
                    line_ele[-1] = line_num
                    line_lists[record].append(line_ele)
                else:
                    line_lists['OTHERS'].append([line[:6].rstrip(),
                                                line[6:-1].rstrip(), line_num])

        dfs = {}
        for r in line_lists.items():
            df = pd.DataFrame(r[1], columns=[c['id'] for c in
                                             pdb_records[r[0]]] + ['line_idx'])
            for c in pdb_records[r[0]]:
                try:
                    df[c['id']] = df[c['id']].astype(c['type'])
                except ValueError:
                    # expect ValueError if float/int columns are empty strings
                    df[c['id']] = pd.Series(np.nan, index=df.index)

            dfs[r[0]] = df

        # issue a warning if no atoms have been loaded
        if len(dfs['ATOM']) == 0:
            warnings.warn('No ATOM entries have been loaded. '
                          'Is the input file/text in the pdb format?')

        return dfs

    def amino3to1(self, record='ATOM',
                  residue_col='residue_name', fillna='?'):
        """Creates 1-letter amino acid codes from DataFrame

        Non-canonical amino-acids are converted as follows:
        ASH (protonated ASP) => D
        CYX (disulfide-bonded CYS) => C
        GLH (protonated GLU) => E
        HID/HIE/HIP (different protonation states of HIS) = H
        HYP (hydroxyproline) => P
        MSE (selenomethionine) => M

        Parameters
        ----------
        record : str, default: 'ATOM'
            Specfies the record DataFrame.
        residue_col : str,  default: 'residue_name'
            Column in `record` DataFrame to look for 3-letter amino acid
            codes for the conversion.
        fillna : str, default: '?'
            Placeholder string to use for unknown amino acids.

        Returns
        ---------
        pandas.DataFrame : Pandas DataFrame object consisting of two columns,
            `'chain_id'` and `'residue_name'`, where the former contains
            the chain ID of the amino acid and the latter
            contains the 1-letter amino acid code, respectively.

        """
        tmp = self.df[record]
        cmp = 'placeholder'
        indices = []

        residue_number_insertion = (tmp['residue_number'].astype(str)
                                    + tmp['insertion'])

        for num, ind in zip(residue_number_insertion, np.arange(tmp.shape[0])):
            if num != cmp:
                indices.append(ind)
            cmp = num

        transl = tmp.iloc[indices][residue_col].map(
            amino3to1dict).fillna(fillna)

        return pd.concat((tmp.iloc[indices]['chain_id'], transl), axis=1)

    def distance(self, xyz=(0.00, 0.00, 0.00), records=('ATOM', 'HETATM')):
        """Computes Euclidean distance between atoms and a 3D point.

        Parameters
        ----------
        xyz : tuple, default: (0.00, 0.00, 0.00)
            X, Y, and Z coordinate of the reference center for the distance
            computation.
        records : iterable, default: ('ATOM', 'HETATM')
            Specify which record sections to consider. For example, to consider
            both protein and ligand atoms, set `records=('ATOM', 'HETATM')`.
            This setting is ignored if `df` is not set to None.
            For downward compatibility, a string argument is still supported
            but deprecated and will be removed in future versions.

        Returns
        ---------
        pandas.Series : Pandas Series object containing the Euclidean
            distance between the atoms in the record section and `xyz`.

        """

        if isinstance(records, str):
            warnings.warn('Using a string as `records` argument is '
                          'deprecated and will not be supported in future'
                          ' versions. Please use a tuple or'
                          ' other iterable instead', DeprecationWarning)
            records = (records,)

        df = pd.concat(objs=[self.df[i] for i in records])

        return np.sqrt(np.sum(df[[
            'x_coord', 'y_coord', 'z_coord']]
            .subtract(xyz, axis=1)**2, axis=1))

    @staticmethod
    def distance_df(df, xyz=(0.00, 0.00, 0.00)):
        """Computes Euclidean distance between atoms and a 3D point.

        Parameters
        ----------
        df : DataFrame
            DataFrame containing entries in the `PandasPdb.df['ATOM']`
            or `PandasPdb.df['HETATM']` format for the
            the distance computation to the `xyz` reference coordinates.
        xyz : tuple, default: (0.00, 0.00, 0.00)
            X, Y, and Z coordinate of the reference center for the distance
            computation.

        Returns
        ---------
        pandas.Series : Pandas Series object containing the Euclidean
            distance between the atoms in the record section and `xyz`.

        """
        return np.sqrt(np.sum(df[[
            'x_coord', 'y_coord', 'z_coord']]
            .subtract(xyz, axis=1)**2, axis=1))

    def to_pdb(self, path, records=None, gz=False, append_newline=True):
        """Write record DataFrames to a PDB file or gzipped PDB file.

        Parameters
        ----------
        path : str
            A valid output path for the pdb file

        records : iterable, default: None
            A list of PDB record sections in
            {'ATOM', 'HETATM', 'ANISOU', 'OTHERS'} that are to be written.
            Writes all lines to PDB if `records=None`.

        gz : bool, default: False
            Writes a gzipped PDB file if True.

        append_newline : bool, default: True
            Appends a new line at the end of the PDB file if True

        """
        if gz:
            openf = gzip.open
            w_mode = 'wt'
        else:
            openf = open
            w_mode = 'w'
        if not records:
            records = self.df.keys()

        dfs = {r: self.df[r].copy() for r in records if not self.df[r].empty}

        for r in dfs.keys():
            for col in pdb_records[r]:
                dfs[r][col['id']] = dfs[r][col['id']].apply(col['strf'])
                dfs[r]['OUT'] = pd.Series('', index=dfs[r].index)

            for c in dfs[r].columns:
                if c in {'line_idx', 'OUT'}:
                    pass
                elif r in {'ATOM', 'HETATM'} and c not in pdb_df_columns:
                    warn('Column %s is not an expected column and'
                         ' will be skipped.' % c)
                else:
                    dfs[r]['OUT'] = dfs[r]['OUT'] + dfs[r][c]

        if pd_version < LooseVersion('0.17.0'):
            warn("You are using an old pandas version (< 0.17)"
                 " that relies on the old sorting syntax."
                 " Please consider updating your pandas"
                 " installation to a more recent version.",
                 DeprecationWarning)
            dfs.sort(columns='line_idx', inplace=True)

        elif pd_version < LooseVersion('0.23.0'):
            df = pd.concat(dfs)

        else:
            df = pd.concat(dfs, sort=False)

        df.sort_values(by='line_idx', inplace=True)

        with openf(path, w_mode) as f:

            s = df['OUT'].tolist()
            for idx in range(len(s)):
                if len(s[idx]) < 80:
                    s[idx] = '%s%s' % (s[idx], ' ' * (80 - len(s[idx])))
            to_write = '\n'.join(s)
            f.write(to_write)
            if append_newline:
                if gz:
                    f.write('\n')
                else:
                    f.write('\n')

    def parse_sse(self):
        """Parse secondary structure elements"""



""" Class for working with Tripos MOL2 files"""
# BioPandas
# Author: Sebastian Raschka <mail@sebastianraschka.com>
# License: BSD 3 clause
# Project Website: http://rasbt.github.io/biopandas/
# Code Repository: https://github.com/rasbt/biopandas

import pandas as pd
import numpy as np
from .mol2_io import split_multimol2


COLUMN_NAMES = (
 'atom_id',
 'atom_name',
 'x',
 'y',
 'z',
 'atom_type',
 'subst_id',
 'subst_name',
 'charge'
)

COLUMN_TYPES = (int, str, float, float, float, str, int, str, float)


class PandasMol2(object):
    """
    Object for working with Tripos Mol2 structure files.

    Attributes
    ----------
    df : pandas.DataFrame
        DataFrame of a Mol2's ATOM section

    mol2_text : str
        Mol2 file contents in string format

    code : str
        ID, code, or name of the molecule stored

    pdb_path : str
        Location of the MOL2 file that was read in via `read_mol2`

    """
    def __init__(self):
        self._df = None
        self.mol2_text = ''
        self.header = ''
        self.code = ''
        self.mol2_path = ''

    @property
    def df(self):
        """Acccesses the pandas DataFrame"""
        return self._df

    @df.setter
    def df(self, value):
        """Assign a new value to the pandas DataFrame"""
        raise AttributeError('Please use `PandasMol2._df = ... ` instead\n'
                             'of `PandasMol2.df = ... ` if you are sure that\n'
                             'you want to overwrite the `df` attribute.')
        # self._df = value

    def _load_mol2(self, mol2_lines, mol2_code, columns):
        """Load mol2 contents into assert_raise_message instance"""
        if columns is None:
            col_names = COLUMN_NAMES
            col_types = COLUMN_TYPES
        else:
            col_names, col_types = [], []
            for i in range(len(columns)):
                col_names.append(columns[i][0])
                col_types.append(columns[i][1])

        try:
            self.mol2_text = ''.join(mol2_lines)
            self.code = mol2_code
        except TypeError:
            mol2_lines = [m.decode() for m in mol2_lines]
            self.mol2_text = ''.join(mol2_lines)
            self.code = mol2_code.decode()

        self._df = self._construct_df(mol2_lines, col_names, col_types)

    def read_mol2(self, path, columns=None):
        """Reads Mol2 files (unzipped or gzipped) from local drive

        Note that if your mol2 file contains more than one molecule,
        only the first molecule is loaded into the DataFrame

        Attributes
        ----------
        path : str
            Path to the Mol2 file in .mol2 format or gzipped format (.mol2.gz)

        columns : dict or None (default: None)
            If None, this methods expects a 9-column ATOM section that contains
            the following columns:

            {0:('atom_id', int), 1:('atom_name', str),
             2:('x', float), 3:('y', float), 4:('z', float),
             5:('atom_type', str), 6:('subst_id', int),
             7:('subst_name', str), 8:('charge', float)}

            If your Mol2 files are formatted differently, you can provide your
            own column_mapping dictionary in a format similar to the one above.
            However, note that not all assert_raise_message methods
            may be supported then.

        Returns
        ---------
        self

        """
        mol2_code, mol2_lines = next(split_multimol2(path))
        self._load_mol2(mol2_lines, mol2_code, columns)
        self.mol2_path = path
        return self

    def read_mol2_from_list(self, mol2_lines, mol2_code, columns=None):
        r"""Reads Mol2 file from a list into DataFrames

        Attributes
        ----------
        mol2_lines : list
            A list of lines containing the mol2 file contents. For example,
            ['@<TRIPOS>MOLECULE\n',
             'ZINC38611810\n',
             '   65    68     0     0     0\n',
             'SMALL\n',
             'NO_CHARGES\n',
             '\n',
             '@<TRIPOS>ATOM\n',
             '      1 C1  -1.1786  2.7011  -4.0323 C.3  1 <0>   -0.1537\n',
             '      2 C2  -1.2950  1.2442  -3.5798 C.3  1 <0>   -0.1156\n',
             ...]

        mol2_code : str or None
            Name or ID of the molecule.

        columns : dict or None (default: None)
            If None, this methods expects a 9-column ATOM section that contains
            the following columns:
            {0:('atom_id', int), 1:('atom_name', str),
             2:('x', float), 3:('y', float), 4:('z', float),
             5:('atom_type', str), 6:('subst_id', int),
             7:('subst_name', str), 8:('charge', float)}
            If your Mol2 files are formatted differently, you can provide your
            own column_mapping dictionary in a format similar to the one above.
            However, note that not all assert_raise_message methods may be
            supported then.

        Returns
        ---------
        self

        """
        self._load_mol2(mol2_lines, mol2_code, columns)
        return self

    def _construct_df(self, mol2_lines, col_names, col_types):
        """Construct DataFrames from list of PDB lines."""
        return self._atomsection_to_pandas(self._get_atomsection(mol2_lines),
                                           col_names=col_names,
                                           col_types=col_types)

    @staticmethod
    def _get_atomsection(mol2_lst):
        """Returns atom section from mol2 provided as list of strings.
        Raises ValueError if data is not provided in the mol2 format."""
        started = False
        first_idx = None
        for idx, s in enumerate(mol2_lst):
            if s.startswith('@<TRIPOS>ATOM'):
                first_idx = idx + 1
                started = True
            elif started and s.startswith('@<TRIPOS>'):
                last_idx_plus1 = idx
                break
        if first_idx is None:
            # Raise error when file contains no @<TRIPOS>ATOM
            # (i.e. file is no mol2 file)
            raise ValueError(
                    "Structural data could not be loaded. "
                    "Is the input file/text in the mol2 format?"
                )
        return mol2_lst[first_idx:last_idx_plus1]

    @staticmethod
    def _atomsection_to_pandas(mol2_atom_lst, col_names, col_types):

        df = pd.DataFrame([lst.split() for lst in mol2_atom_lst],
                          columns=col_names)

        for i in range(df.shape[1]):
            df[col_names[i]] = df[col_names[i]].astype(col_types[i])

        return df

    @staticmethod
    def rmsd(df1, df2, heavy_only=True):
        """Compute the Root Mean Square Deviation between molecules

        Parameters
        ----------
        df1 : pandas.DataFrame
            DataFrame with HETATM, ATOM, and/or ANISOU entries
        df2 : pandas.DataFrame
            Second DataFrame for RMSD computation against df1. Must have the
            same number of entries as df1
        heavy_only : bool (default: True)
            Which atoms to compare to compute the RMSD. If `True` (default),
            computes the RMSD between non-hydrogen atoms only.

        Returns
        ---------
        rmsd : float
            Root Mean Square Deviation between df1 and df2

        """
        if df1.shape[0] != df2.shape[0]:
            raise AttributeError('DataFrames have unequal lengths')

        if heavy_only:
            d1 = df1[df1['atom_type'] != 'H']
            d2 = df2[df2['atom_type'] != 'H']
        else:
            d1, d2 = df1, df2

        total = ((d1['x'].values - d2['x'].values)**2 +
                 (d1['y'].values - d2['y'].values)**2 +
                 (d1['z'].values - d2['z'].values)**2)
        rmsd = round((total.sum() / df1.shape[0])**0.5, 4)
        return rmsd

    def distance(self, xyz=(0.00, 0.00, 0.00)):
        """Computes Euclidean distance between atoms in
            self.df and a 3D point.

        Parameters
        ----------
        xyz : tuple (0.00, 0.00, 0.00)
            X, Y, and Z coordinate of the reference center for the distance
            computation

        Returns
        ---------
        pandas.Series : Pandas Series object containing the Euclidean
            distance between the atoms in the atom section and `xyz`.

        """
        return np.sqrt(np.sum(self.df[['x', 'y', 'z']]
                       .subtract(xyz, axis=1)**2, axis=1))

    @staticmethod
    def distance_df(df, xyz=(0.00, 0.00, 0.00)):
        """Computes Euclidean distance between atoms and a 3D point.

        Parameters
        ----------
        df : DataFrame
            DataFrame containing entries similar to the PandasMol2.df
            format for the
            the distance computation to the `xyz` reference coordinates.
        xyz : tuple (0.00, 0.00, 0.00)
            X, Y, and Z coordinate of the reference center for the distance
            computation

        Returns
        ---------
        pandas.Series : Pandas Series object containing the Euclidean
            distance between the atoms in the atom section and `xyz`.

        """

        return np.sqrt(np.sum(df[['x', 'y', 'z']]
                       .subtract(xyz, axis=1)**2, axis=1))



# BioPandas
# Author: Sebastian Raschka <mail@sebastianraschka.com>
# License: BSD 3 clause
# Project Website: http://rasbt.github.io/biopandas/
# Code Repository: https://github.com/rasbt/biopandas

import pandas as pd
from biopandas.pdb import PandasPdb
import os


def test_equal():
    TESTDATA_1t48 = os.path.join(os.path.dirname(__file__), 'data',
                                                            '1t48_995.pdb')

    p1t48 = PandasPdb()
    p1t48.read_pdb(TESTDATA_1t48)
    dist = p1t48.distance(xyz=(70.785, 15.477, 23.359), records=('ATOM',))

    expect = pd.Series([2.533259, 1.520502, 0.000000, 1.257597, 1.252510],
                       index=[12, 13, 14, 15, 16])
    assert dist[dist < 3].all() == expect.all()


def test_deprecated_str_arg():
    TESTDATA_1t48 = os.path.join(os.path.dirname(__file__), 'data',
                                                            '1t48_995.pdb')

    p1t48 = PandasPdb()
    p1t48.read_pdb(TESTDATA_1t48)
    dist = p1t48.distance(xyz=(70.785, 15.477, 23.359), records='ATOM')

    expect = pd.Series([2.533259, 1.520502, 0.000000, 1.257597, 1.252510],
                       index=[12, 13, 14, 15, 16])
    assert dist[dist < 3].all() == expect.all()


def test_use_external_df():
    TESTDATA_1t48 = os.path.join(os.path.dirname(__file__), 'data',
                                                            '1t48_995.pdb')

    p1t48 = PandasPdb()
    p1t48.read_pdb(TESTDATA_1t48)
    new_df = p1t48.df['ATOM'].iloc[:-1, :].copy()
    dist = PandasPdb.distance_df(df=new_df, xyz=(70.785, 15.477, 23.359))

    expect = pd.Series([2.533259, 1.520502, 0.000000, 1.257597],
                       index=[12, 13, 14, 15])
    assert dist[dist < 3].all() == expect.all()



# BioPandas
# Author: Sebastian Raschka <mail@sebastianraschka.com>
# License: BSD 3 clause
# Project Website: http://rasbt.github.io/biopandas/
# Code Repository: https://github.com/rasbt/biopandas


from biopandas.pdb import PandasPdb
import os
import numpy as np
import pandas as pd
from nose.tools import raises

TESTDATA_FILENAME = os.path.join(os.path.dirname(__file__), 'data',
                                 '3eiy_stripped_no_ele.pdb')

ppdb = PandasPdb()
ppdb.read_pdb(TESTDATA_FILENAME)


def test_impute_hetatm():
    new = ppdb.impute_element(records=['HETATM'])
    assert new['HETATM']['element_symbol'][1] == 'N'
    assert new['HETATM']['element_symbol'][10] == 'O'
    assert new['ATOM']['element_symbol'][1] == ''
    assert new['ATOM']['element_symbol'][10] == ''


def test_impute_atom():
    new = ppdb.impute_element(records=['ATOM'])
    assert new['ATOM']['element_symbol'][1] == 'C'
    assert new['ATOM']['element_symbol'][10] == 'C'
    assert new['HETATM']['element_symbol'][1] == ''
    assert new['HETATM']['element_symbol'][10] == ''


def test_impute_default():
    assert ppdb.df['ATOM']['element_symbol'][1] == ''
    assert ppdb.df['HETATM']['element_symbol'][1] == ''
    new = ppdb.impute_element()
    assert new['ATOM']['element_symbol'][1] == 'C'
    assert new['ATOM']['element_symbol'][10] == 'C'
    assert new['HETATM']['element_symbol'][1] == 'N'
    assert new['HETATM']['element_symbol'][10] == 'O'



# BioPandas
# Author: Sebastian Raschka <mail@sebastianraschka.com>
# License: BSD 3 clause
# Project Website: http://rasbt.github.io/biopandas/
# Code Repository: https://github.com/rasbt/biopandas

import difflib


def assertMultiLineEqual(first, second, preserve_newline=True, msg=None):
    """Assert that two multi-line strings are equal."""
    assert isinstance(first, str), 'First argument is not a string'
    assert isinstance(second, str), 'Second argument is not a string'

    if first != second:
        message = ''.join(difflib.ndiff(first.splitlines(preserve_newline),
                                        second.splitlines(preserve_newline)))
        if msg:
            message += " : " + msg
        raise AssertionError("Multi-line strings are unequal:\n" + message)


def assert_raises(exception_type, message, func, *args, **kwargs):
    """Check that an exception is raised with a specific message

    Parameters
    ----------
    exception_type : exception
        The exception that should be raised
    message : str (default: None)
        The error message that should be raised. Ignored if False or None
    func : callable
        The function that raises the exception
    *args : positional arguments to `func`
    **kwargs : keyword arguments to `func`

    """
    try:
        func(*args, **kwargs)
    except exception_type as e:
        error_message = str(e)
        if message and message not in error_message:
            raise AssertionError("Error message differs from the expected"
                                 " string: %r. Got error message: %r" %
                                 (message, error_message))
    else:
        raise AssertionError('%s not raised.' % exception_type.__name__)



# BioPandas
# Author: Sebastian Raschka <mail@sebastianraschka.com>
# License: BSD 3 clause
# Project Website: http://rasbt.github.io/biopandas/
# Code Repository: https://github.com/rasbt/biopandas

"""
BioPandas module for working with TRIPOS MOL2
files in pandas DataFrames.
"""

from .pandas_mol2 import PandasMol2
from .mol2_io import split_multimol2

__all__ = ["PandasMol2", "split_multimol2"]



# BioPandas
# Author: Sebastian Raschka <mail@sebastianraschka.com>
# License: BSD 3 clause
# Project Website: http://rasbt.github.io/biopandas/
# Code Repository: https://github.com/rasbt/biopandas

from biopandas.pdb import PandasPdb
import os
from nose.tools import raises


TESTDATA_1t48 = os.path.join(os.path.dirname(__file__), 'data',
                                                        '1t48_995.pdb')
TESTDATA_1t49 = os.path.join(os.path.dirname(__file__), 'data',
                                                        '1t49_995.pdb')
TESTDATA_lig1 = os.path.join(os.path.dirname(__file__), 'data',
                                                        'lig_conf_1.pdb')
TESTDATA_lig2 = os.path.join(os.path.dirname(__file__), 'data',
                                                        'lig_conf_2.pdb')

TESTDATA_rna = os.path.join(os.path.dirname(__file__), 'data',
                                                       '1ehz-rna_short.pdb')

p1t48 = PandasPdb()
p1t48.read_pdb(TESTDATA_1t48)
p1t49 = PandasPdb()
p1t49.read_pdb(TESTDATA_1t49)

pl1 = PandasPdb()
pl1.read_pdb(TESTDATA_lig1)
pl2 = PandasPdb()
pl2.read_pdb(TESTDATA_lig2)


def test_equal():
    r = PandasPdb.rmsd(p1t48.df['ATOM'], p1t48.df['ATOM'], s=None)
    assert r == 0.000, r


@raises(AttributeError)
def test_wrong_arg():
    PandasPdb.rmsd(p1t48.df['ATOM'].loc[1:, :], p1t48.df['ATOM'], s='bla')


@raises(AttributeError)
def test_incompatible():
    PandasPdb.rmsd(p1t48.df['ATOM'].loc[1:, :], p1t48.df['ATOM'], s=None)


@raises(AttributeError)
def test_invalid_query():
    PandasPdb.rmsd(p1t48.df['ATOM'].loc[1:, :], p1t48.df['ATOM'], s='bla')


def test_protein():
    r = PandasPdb.rmsd(p1t48.df['ATOM'], p1t49.df['ATOM'],
                       s='c-alpha', invert=False)
    assert r == 0.4785, r


def test_rna_and_nonmatching_indices():
    ehz = PandasPdb().read_pdb(TESTDATA_rna)
    at = ehz.df['ATOM']
    a64 = at[at['residue_number'] == 64]
    a66 = at[at['residue_number'] == 66]
    r = PandasPdb.rmsd(a64, a66)
    assert r == 10.2007, r


def test_ligand():
    r = PandasPdb.rmsd(pl1.df['HETATM'], pl2.df['HETATM'],
                       s='hydrogen', invert=True)
    assert r == 1.9959, r


def test_ligand_default():
    r = PandasPdb.rmsd(pl1.df['HETATM'], pl2.df['HETATM'],
                       s=None)
    assert r == 2.6444, r



# BioPandas
# Author: Sebastian Raschka <mail@sebastianraschka.com>
# License: BSD 3 clause
# Project Website: http://rasbt.github.io/biopandas/
# Code Repository: https://github.com/rasbt/biopandas

import gzip


def split_multimol2(mol2_path):
    r"""
    Generator function that
    splits a multi-mol2 file into individual Mol2 file contents.

    Parameters
    -----------
    mol2_path : str
      Path to the multi-mol2 file. Parses gzip files if the filepath
      ends on .gz.

    Returns
    -----------
    A generator object for lists for every extracted mol2-file. Lists contain
        the molecule ID and the mol2 file contents.
        e.g., ['ID1234', ['@<TRIPOS>MOLECULE\n', '...']]. Note that bytestrings
        are returned (for reasons of efficieny) if the Mol2 content is read
        from a gzip (.gz) file.

    """
    if mol2_path.endswith('.mol2'):
        open_file = open
        read_mode = 'r'
    elif mol2_path.endswith('mol2.gz'):
        open_file = gzip.open
        read_mode = 'rb'
    else:
        raise ValueError('Wrong file format;'
                         'allowed file formats are .mol2 and .mol2.gz.')

    check = {'rb': b'@<TRIPOS>MOLECULE', 'r': '@<TRIPOS>MOLECULE'}

    with open_file(mol2_path, read_mode) as f:
        mol2 = ['', []]
        while True:
            try:
                line = next(f)
                if line.startswith(check[read_mode]):
                    if mol2[0]:
                        yield(mol2)
                    mol2 = ['', []]
                    mol2_id = next(f)
                    mol2[0] = mol2_id.rstrip()
                    mol2[1].append(line)
                    mol2[1].append(mol2_id)
                else:
                    mol2[1].append(line)
            except StopIteration:
                yield(mol2)
                return



# BioPandas
# Author: Sebastian Raschka <mail@sebastianraschka.com>
# License: BSD 3 clause
# Project Website: http://rasbt.github.io/biopandas/
# Code Repository: https://github.com/rasbt/biopandas

from biopandas.pdb import PandasPdb
from biopandas.testutils import assert_raises
import os


TESTDATA_FILENAME = os.path.join(os.path.dirname(__file__), 'data', '3eiy.pdb')


def test_overwrite_df():
    data_path = os.path.join(os.path.dirname(__file__), 'data', '3eiy.pdb')
    pdb = PandasPdb().read_pdb(data_path)

    def overwrite():
        pdb.df = 'bla'

    expect = ('Please use `PandasPdb._df = ... ` instead\n'
              'of `PandasPdb.df = ... ` if you are sure that\n'
              'you want to overwrite the `df` attribute.')

    assert_raises(AttributeError,
                  expect,
                  overwrite)



import io
import os
from setuptools import setup, find_packages

VERSION = None
with io.open(
    os.path.join(os.path.dirname(__file__), 'biopandas/__init__.py'),
    encoding='utf-8'
) as f:
    for l in f:
        if not l.startswith('__version__'):
            continue
        VERSION = l.split('=')[1].strip(' "\'\n')
        break
PROJECT_ROOT = os.path.dirname(os.path.realpath(__file__))

REQUIREMENTS_FILE = os.path.join(PROJECT_ROOT, 'requirements.txt')

with open(REQUIREMENTS_FILE) as f:
    install_reqs = f.read().splitlines()

install_reqs.append('setuptools')

setup(name='biopandas',
      version=VERSION,
      description='Machine Learning Library Extensions',
      author='Sebastian Raschka',
      author_email='mail@sebastianraschka.com',
      url='https://github.com/rasbt/biopandas',
      packages=find_packages(),
      package_data={'': ['LICENSE.txt',
                         'README.md',
                         'requirements.txt']
                    },
      include_package_data=True,
      install_requires=install_reqs,
      license='BSD 3-Clause',
      platforms='any',
      classifiers=[
             'License :: OSI Approved :: BSD License',
             'Development Status :: 5 - Production/Stable',
             'Operating System :: Microsoft :: Windows',
             'Operating System :: POSIX',
             'Operating System :: Unix',
             'Operating System :: MacOS',
             'Programming Language :: Python :: 3',
             'Programming Language :: Python :: 3.5',
             'Programming Language :: Python :: 3.6',
             'Programming Language :: Python :: 3.7',
             'Topic :: Scientific/Engineering',
      ],
      long_description="""
Biopandas is a python package for working with molecular structures
in pandas DataFrames.


Contact
=============

If you have any questions or comments about biopandas,
please feel free to contact me via
eMail: mail@sebastianraschka.com
or Twitter: https://twitter.com/rasbt

This project is hosted at https://github.com/rasbt/biopandas

The documentation can be found at http://rasbt.github.io/biopandas/

""")



# BioPandas
# Author: Sebastian Raschka <mail@sebastianraschka.com>
# License: BSD 3 clause
# Project Website: http://rasbt.github.io/biopandas/
# Code Repository: https://github.com/rasbt/biopandas
#
#
#
#
# PEP0440 compatible formatted version, see:
# https://www.python.org/dev/peps/pep-0440/
#
# Generic release markers:
#   X.Y
#   X.Y.Z   # For bugfix releases
#
# Admissible pre-release markers:
#   X.YaN   # Alpha release
#   X.YbN   # Beta release
#   X.YrcN  # Release Candidate
#   X.Y     # Final release
#
# Dev branch marker is: 'X.Y.dev' or 'X.Y.devN' where N is an integer.
# 'X.Y.dev0' is the canonical version of 'X.Y.dev'
#

__version__ = '0.2.8'
__author__ = "Sebastian Raschka <mail@sebastianraschka.com>"
